/**
 * The MIT License (MIT)
 * Copyright (c) 2017-present Dmitry Soshnikov <dmitry.soshnikov@gmail.com>
 */
'use strict';

var DFA = require('../dfa');

var NFA = require('../../nfa/nfa');

var NFAState = require('../../nfa/nfa-state');

var _require = require('../../nfa/builders'),
    char = _require.char,
    rep = _require.rep;

var _require2 = require('../../special-symbols'),
    EPSILON = _require2.EPSILON; // x|y


function getDefaultNFA() {
  var A = new NFAState();
  var B = new NFAState();
  var C = new NFAState();
  var D = new NFAState();
  var E = new NFAState();
  var F = new NFAState({
    accepting: true
  }); // x|y

  A.addTransition(EPSILON, B);
  A.addTransition(EPSILON, C); // x

  B.addTransition('x', D); // y

  C.addTransition('y', E);
  D.addTransition(EPSILON, F);
  E.addTransition(EPSILON, F);
  return new NFA(A, F);
}

describe('dfa', function () {
  it('alphabet', function () {
    var dfa = new DFA(getDefaultNFA());
    expect(dfa.getAlphabet()).toEqual(new Set(['x', 'y']));
  });
  it('accepting states', function () {
    var dfa = new DFA(getDefaultNFA());
    expect(dfa.getOriginaAcceptingStateNumbers()).toEqual(new Set(['6,4', '3,4']));
    expect(dfa.getAcceptingStateNumbers()).toEqual(new Set([2, 3]));
  });
  it('transition table', function () {
    var dfa = new DFA(getDefaultNFA());
    expect(dfa.getOriginalTransitionTable()).toEqual({
      '1,2,5': {
        'x': '3,4',
        'y': '6,4'
      },
      '3,4': {},
      '6,4': {}
    });
    expect(dfa.getTransitionTable()).toEqual({
      1: {
        'x': 3,
        'y': 2
      },
      2: {},
      3: {}
    });
  });
  it('matches', function () {
    var dfa = new DFA(getDefaultNFA());
    expect(dfa.matches('x')).toBe(true);
    expect(dfa.matches('y')).toBe(true);
    expect(dfa.matches('z')).toBe(false);
    expect(dfa.matches('')).toBe(false);
  });
  it('matches rep', function () {
    var dfa = new DFA(rep(char('a')));
    expect(dfa.matches('')).toBe(true);
    expect(dfa.matches('a')).toBe(true);
    expect(dfa.matches('aa')).toBe(true);
    expect(dfa.matches('aaa')).toBe(true);
  });
  it('minimizes', function () {
    var dfa = new DFA(getDefaultNFA());
    expect(dfa.getTransitionTable()).toEqual({
      1: {
        'x': 3,
        'y': 2
      },
      2: {},
      3: {}
    });
    dfa.minimize();
    expect(dfa.getTransitionTable()).toEqual({
      1: {
        'x': 2,
        'y': 2
      },
      2: {}
    });
  });
  it('matches minimize', function () {
    var dfa = new DFA(getDefaultNFA());
    dfa.minimize();
    expect(dfa.matches('x')).toBe(true);
    expect(dfa.matches('y')).toBe(true);
    expect(dfa.matches('z')).toBe(false);
    expect(dfa.matches('')).toBe(false);
  });
  it('matches rep', function () {
    var dfa = new DFA(rep(char('a')));
    dfa.minimize();
    expect(dfa.matches('')).toBe(true);
    expect(dfa.matches('a')).toBe(true);
    expect(dfa.matches('aa')).toBe(true);
    expect(dfa.matches('aaa')).toBe(true);
  });
});