/**
 * The MIT License (MIT)
 * Copyright (c) 2017-present Dmitry Soshnikov <dmitry.soshnikov@gmail.com>
 */
'use strict';

function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread(); }

function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance"); }

function _iterableToArray(iter) { if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter); }

function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } }

var NFA = require('../nfa');

var NFAState = require('../nfa-state');

var _require = require('../../special-symbols'),
    EPSILON = _require.EPSILON;

var _require2 = require('../builders'),
    alt = _require2.alt,
    char = _require2.char,
    e = _require2.e,
    or = _require2.or,
    rep = _require2.rep,
    repExplicit = _require2.repExplicit,
    plusRep = _require2.plusRep,
    questionRep = _require2.questionRep;

function getAtIndex(set, index) {
  return _toConsumableArray(set)[index];
}

describe('nfa-builders', function () {
  it('char', function () {
    var a = char('a');
    expect(a).toBeInstanceOf(NFA);
    expect(a.in).toBeInstanceOf(NFAState);
    expect(a.out).toBeInstanceOf(NFAState);
    expect(a.in.accepting).toBe(false);
    expect(a.out.accepting).toBe(true);
    expect(a.in.getTransitionsOnSymbol('a').size).toBe(1);
    expect(a.in.getTransitionsOnSymbol('a')).toEqual(new Set([a.out]));
    expect(a.matches('a')).toBe(true);
  });
  it('e', function () {
    var epsilon = e();
    expect(epsilon).toBeInstanceOf(NFA);
    expect(epsilon.in).toBeInstanceOf(NFAState);
    expect(epsilon.out).toBeInstanceOf(NFAState);
    expect(epsilon.in.accepting).toBe(false);
    expect(epsilon.out.accepting).toBe(true);
    expect(epsilon.in.getTransitionsOnSymbol(EPSILON).size).toBe(1);
    expect(epsilon.in.getTransitionsOnSymbol(EPSILON)).toEqual(new Set([epsilon.out]));
    expect(epsilon.matches('')).toBe(true);
  });
  it('or', function () {
    var a = char('a');
    var b = char('b');
    var c = char('c'); // Before patching chars are accepting.

    expect(a.out.accepting).toBe(true);
    expect(b.out.accepting).toBe(true);
    expect(c.out.accepting).toBe(true);
    var AorBorC = or(a, b, c);
    expect(AorBorC).toBeInstanceOf(NFA); // After patching chars out are not accepting.

    expect(a.out.accepting).toBe(false);
    expect(b.out.accepting).toBe(false);
    expect(c.out.accepting).toBe(false);
    expect(AorBorC.out.accepting).toBe(true);
    var AorB = getAtIndex(AorBorC.in.getTransitionsOnSymbol(EPSILON), 0);
    var partA = getAtIndex(AorB.getTransitionsOnSymbol(EPSILON), 0);
    var partB = getAtIndex(AorB.getTransitionsOnSymbol(EPSILON), 1);
    expect(partA).toBe(a.in);
    expect(partB).toBe(b.in);
    var partC = getAtIndex(AorBorC.in.getTransitionsOnSymbol(EPSILON), 1);
    expect(partC).toBe(c.in);
    var outFromB = getAtIndex(getAtIndex(b.out.getTransitionsOnSymbol(EPSILON), 0).getTransitionsOnSymbol(EPSILON), 0);
    expect(outFromB).toBe(AorBorC.out);
    var outFromC = getAtIndex(c.out.getTransitionsOnSymbol(EPSILON), 0);
    expect(outFromC).toBe(AorBorC.out);
    expect(AorBorC.matches('a')).toBe(true);
    expect(AorBorC.matches('b')).toBe(true);
    expect(AorBorC.matches('c')).toBe(true);
  });
  it('alt', function () {
    var a = char('a');
    var b = char('b');
    var c = char('c'); // Before patching chars are accepting.

    expect(a.out.accepting).toBe(true);
    expect(b.out.accepting).toBe(true);
    expect(c.out.accepting).toBe(true);
    var ABC = alt(a, b, c);
    expect(ABC).toBeInstanceOf(NFA); // After patching chars out are not accepting.

    expect(a.out.accepting).toBe(false);
    expect(b.out.accepting).toBe(false);
    expect(c.out.accepting).toBe(true);
    expect(ABC.in).toBe(a.in);
    expect(ABC.out).toBe(c.out);
    expect(getAtIndex(a.out.getTransitionsOnSymbol(EPSILON), 0)).toBe(b.in);
    expect(getAtIndex(b.out.getTransitionsOnSymbol(EPSILON), 0)).toBe(c.in);
    expect(ABC.matches('abc')).toBe(true);
  });
  it('kleene-closure-explicit', function () {
    var a = char('a');
    var aRep = repExplicit(a);
    expect(aRep).toBeInstanceOf(NFA);
    var partA = getAtIndex(aRep.in.getTransitionsOnSymbol(EPSILON), 0);
    expect(partA).toBe(a.in);
    var partEpsilon = getAtIndex(aRep.in.getTransitionsOnSymbol(EPSILON), 1);
    expect(partEpsilon.accepting).toBe(true);
    expect(partEpsilon).toBe(aRep.out);
    var backToA = getAtIndex(aRep.out.getTransitionsOnSymbol(EPSILON), 0);
    expect(backToA).toBe(a.in);
    expect(aRep.matches('')).toBe(true);
    expect(aRep.matches('a')).toBe(true);
    expect(aRep.matches('aaa')).toBe(true);
  });
  it('kleene-closure', function () {
    var a = char('a');
    var aRep = rep(a);
    expect(aRep).toBe(a);
    expect(aRep).toBeInstanceOf(NFA);
    var partEpsilon = getAtIndex(aRep.in.getTransitionsOnSymbol(EPSILON), 0);
    expect(partEpsilon.accepting).toBe(true);
    expect(partEpsilon).toBe(aRep.out);
    var backToA = getAtIndex(aRep.out.getTransitionsOnSymbol(EPSILON), 0);
    expect(backToA).toBe(a.in);
    expect(aRep.matches('')).toBe(true);
    expect(aRep.matches('a')).toBe(true);
    expect(aRep.matches('aaa')).toBe(true);
  });
  it('plusRep', function () {
    var a = char('a');
    var aPlus = plusRep(a);
    expect(aPlus).toBe(a);
    expect(aPlus).toBeInstanceOf(NFA);
    var backToA = getAtIndex(aPlus.out.getTransitionsOnSymbol(EPSILON), 0);
    expect(backToA).toBe(a.in);
    expect(aPlus.matches('')).toBe(false);
    expect(aPlus.matches('a')).toBe(true);
    expect(aPlus.matches('aaa')).toBe(true);
  });
  it('questionRep', function () {
    var a = char('a');
    var aQuestion = questionRep(a);
    expect(aQuestion).toBe(a);
    expect(aQuestion).toBeInstanceOf(NFA);
    var partEpsilon = getAtIndex(aQuestion.in.getTransitionsOnSymbol(EPSILON), 0);
    expect(partEpsilon).toBe(a.out);
    expect(aQuestion.matches('')).toBe(true);
    expect(aQuestion.matches('a')).toBe(true);
    expect(aQuestion.matches('aa')).toBe(false);
  });
});