/**
 * The MIT License (MIT)
 * Copyright (c) 2017-present Dmitry Soshnikov <dmitry.soshnikov@gmail.com>
 */
'use strict';

var optimizer = require('..');

describe('optimizer-integration-test', function () {
  it('optimizes a regexp', function () {
    var original = /[0-90-9a-zA-Z_$][0-9a-zA-Za-z_$]*/;
    var optimized = /[\w$]+/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('preserves dash', function () {
    var original = '/[^a-zа-яё -]+gi/';
    var optimized = '/[^ a-zа-яё-]+gi/';
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
    original = '/[0-9\\-a-z]/';
    optimized = '/[\\da-z\\-]/';
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
    original = /^[a-z][a-z0-9\-]{5,29}$/;
    optimized = /^[a-z][\da-z\-]{5,29}$/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
    original = /[-\da-z]/;
    optimized = /[\da-z-]/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('handles a named capturing group', function () {
    var original = '/(?<any>.*)/';
    var optimized = '/(?<any>.*)/';
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('to single char', function () {
    var original = /[0-90-9a-zA-Z_][0-9a-zA-Za-z_]*/;
    var optimized = /\w+/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('preserve escape', function () {
    var original = /^[\^\*\$\(\)]\^\*\$\(\)\.\[\/\\$/;
    var optimized = /^[$()*^]\^\*\$\(\)\.\[\/\\$/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('whitespace', function () {
    var original = /[ \f\n\r\t\v\u00a0\u1680\u2000-\u200a\u2028\u2029\u202f\u205f\u3000\ufeff]+/;
    var optimized = /\s+/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('whitespace group', function () {
    var original = /[ \f\n\r\t\v\u00a0\u1680\u2000-\u200a\u2028\u2029\u202f\u205f\u3000\ufeff\]]/g;
    var optimized = /[\s\]]/g;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('quantifier {1,}', function () {
    var original = /[a-z0-9A-Za-z_]{1,}/;
    var optimized = /\w+/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('quantifier {1}', function () {
    var original = /[a-z0-9A-Za-z_]{1}/;
    var optimized = /\w/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('quantifier {3,3}', function () {
    var original = /[a-z0-9A-Za-z_]{3,3}/;
    var optimized = /\w{3}/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('quantifier other', function () {
    var original = /[a-z0-9A-Za-z_]{1,3}/;
    var optimized = /\w{1,3}/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('finds the best optimization', function () {
    var original = /a|a/;
    var optimized = /a/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('applies whitelist only', function () {
    var original = '/(?:[a])/';
    var optimized = '/[a]/';
    expect(optimizer.optimize(original, {
      whitelist: ['ungroup']
    }).toString()).toBe(optimized.toString());
  });
  it('sorts characters', function () {
    var original = /[åä]/;
    var optimized = /[äå]/;
    expect(optimizer.optimize(original).toString()).toBe(optimized.toString());
  });
  it('preserves character order when blacklisted charClassClassrangesMerge', function () {
    var original = /[åä]/;
    var optimized = /[åä]/;
    expect(optimizer.optimize(original, {
      blacklist: ['charClassClassrangesMerge']
    }).toString()).toBe(optimized.toString());
  });
});