/**
 * The MIT License (MIT)
 * Copyright (c) 2017-present Dmitry Soshnikov <dmitry.soshnikov@gmail.com>
 */
'use strict';

var _require = require('../../../transform'),
    transform = _require.transform;

var charUnescape = require('../char-escape-unescape-transform');

describe('e -> e', function () {
  it('simple chars', function () {
    var re = transform(/\e\*\]/, [charUnescape]);
    expect(re.toString()).toBe(/e\*]/.toString());
  });
  it('preserve escape', function () {
    var re = transform(/\*\^\$\(\)\[\|/, [charUnescape]);
    expect(re.toString()).toBe(/\*\^\$\(\)\[\|/.toString());
  });
  it('unescapes curly braces', function () {
    var re = transform(/\{\}/, [charUnescape]);
    expect(re.toString()).toBe(/{}/.toString());
  });
  it('preserves dash', function () {
    var re = transform('/[^a-zа-яё -]+gi/', [charUnescape]);
    expect(re.toString()).toBe('/[^a-zа-яё -]+gi/');
    re = transform('/[0-9-a-z]/', [charUnescape]);
    expect(re.toString()).toBe('/[0-9-a-z]/');
    re = transform(/^[a-z][a-z0-9\-]{5,29}$/, [charUnescape]);
    expect(re.toString()).toBe('/^[a-z][a-z0-9\\-]{5,29}$/');
  });
  it('unescapes curly in parent with no index', function () {
    var re = transform(/(?:\{)/, [charUnescape]);
    expect(re.toString()).toBe(/(?:{)/.toString());
    re = transform(/(?:\})/, [charUnescape]);
    expect(re.toString()).toBe(/(?:})/.toString());
    re = transform(/\{/, [charUnescape]);
    expect(re.toString()).toBe(/{/.toString());
    re = transform(/\}/, [charUnescape]);
    expect(re.toString()).toBe(/}/.toString());
    re = transform(/\{|v/, [charUnescape]);
    expect(re.toString()).toBe(/{|v/.toString());
    re = transform(/\}|v/, [charUnescape]);
    expect(re.toString()).toBe(/}|v/.toString());
  });
  it('does not unescape { when looking like a quantifier', function () {
    var re = transform(/a\{3}/, [charUnescape]);
    expect(re.toString()).toBe(/a\{3}/.toString());
    re = transform(/a\{3,}/, [charUnescape]);
    expect(re.toString()).toBe(/a\{3,}/.toString());
    re = transform(/a\{10,12}/, [charUnescape]);
    expect(re.toString()).toBe(/a\{10,12}/.toString());
  });
  it('does not unescape } when looking like a quantifier', function () {
    var re = transform(/a{3\}/, [charUnescape]);
    expect(re.toString()).toBe(/a{3\}/.toString());
    re = transform(/a{3,\}/, [charUnescape]);
    expect(re.toString()).toBe(/a{3,\}/.toString());
    re = transform(/a{10,12\}/, [charUnescape]);
    expect(re.toString()).toBe(/a{10,12\}/.toString());
  });
  it('char class', function () {
    var re = transform(/[\e\*\(\]\ \^\$\/-\?\-]\(\n/, [charUnescape]); // Can't use native toString() conversion on a regexp here
    // because Node 6 stringifies /[/]/ as /[\/]/

    expect(re.toString()).toBe('/[e*(\\] ^$/-?\\-]\\(\\n/');
  });
  it('does not unescape ^ in char class when in first position', function () {
    var re = transform(/[\^a]/, [charUnescape]);
    expect(re.toString()).toBe(/[\^a]/.toString());
  });
  it('does not unescape - in char class when not in first or last position', function () {
    var re = transform(/[a\-z]/, [charUnescape]);
    expect(re.toString()).toBe(/[a\-z]/.toString());
  });
  it('does not unescape space and # when x flag is set', function () {
    var re = transform('/\\ \\#[\\ \\#]/x', [charUnescape]);
    expect(re.toString()).toBe('/\\ \\#[ #]/x');
  });
});