/**
 * The MIT License (MIT)
 * Copyright (c) 2017-present Dmitry Soshnikov <dmitry.soshnikov@gmail.com>
 */
'use strict';

var _require = require('../../../transform'),
    transform = _require.transform;

var ungroup = require('../ungroup-transform');

describe('ungroup', function () {
  it('ungroups simple groups', function () {
    var re = transform(/(?:a)/, [ungroup]);
    expect(re.toString()).toBe('/a/');
    var re2 = transform(/a(?:[\da]b)c/, [ungroup]);
    expect(re2.toString()).toBe('/a[\\da]bc/');
  });
  it('ungroups disjunctions if top-level', function () {
    var re = transform(/(?:ab|bc)/, [ungroup]);
    expect(re.toString()).toBe('/ab|bc/');
  });
  it('does not ungroup other disjunctions', function () {
    var re = transform(/^(?:ab|bc)/, [ungroup]);
    expect(re.toString()).toBe('/^(?:ab|bc)/');
  });
  it('ungroups groups with quantifier if child is a Char', function () {
    var re = transform(/(?:a)+/, [ungroup]);
    expect(re.toString()).toBe('/a+/');
  });
  it('ungroups groups with quantifier if child is a CharacterClass', function () {
    var re = transform(/(?:[ab])+/, [ungroup]);
    expect(re.toString()).toBe('/[ab]+/');
  });
  it('merges Group content of type Alternative with parent Alternative', function () {
    var re = transform(/a(?:bc)d/, [ungroup]);
    expect(re.toString()).toBe('/abcd/');
    var ast = re.getAST();
    expect(ast.body.type === 'Alternative' && ast.body.expressions.every(function (expression) {
      return expression.type === 'Char';
    })).toBe(true);
  });
  it('does not ungroup groups with quantifier otherwise', function () {
    var re = transform(/(?:ab)+/, [ungroup]);
    expect(re.toString()).toBe('/(?:ab)+/');
  });
  it('does not ungroup capturing groups', function () {
    var re = transform(/(a)/, [ungroup]);
    expect(re.toString()).toBe('/(a)/');
  });
  it('does not ungroup empty groups', function () {
    var re = transform(/(?:)+/, [ungroup]);
    expect(re.toString()).toBe('/(?:)+/');
  });
  it('does not extract with reference sibling', function () {
    var re = transform('/(.)\\1(?:0)/', [ungroup]);
    expect(re.toString()).toBe('/(.)\\1(?:0)/');
  });
  it('does not extract with char number sibling', function () {
    var re = transform('/(.)\\2(?:0)/', [ungroup]);
    expect(re.toString()).toBe('/(.)\\2(?:0)/');
  });
});