"use strict";

/**
 * The MIT License (MIT)
 * Copyright (c) 2017-present Dmitry Soshnikov <dmitry.soshnikov@gmail.com>
 */
var parser = require('..');

var _require = require('../unicode/parser-unicode-properties'),
    getCanonicalValue = _require.getCanonicalValue,
    BINARY_PROP_NAMES_TO_ALIASES = _require.BINARY_PROP_NAMES_TO_ALIASES,
    BINARY_ALIASES_TO_PROP_NAMES = _require.BINARY_ALIASES_TO_PROP_NAMES,
    GENERAL_CATEGORY_VALUE_TO_ALIASES = _require.GENERAL_CATEGORY_VALUE_TO_ALIASES,
    GENERAL_CATEGORY_VALUE_ALIASES_TO_VALUES = _require.GENERAL_CATEGORY_VALUE_ALIASES_TO_VALUES,
    SCRIPT_VALUE_TO_ALIASES = _require.SCRIPT_VALUE_TO_ALIASES,
    SCRIPT_VALUE_ALIASES_TO_VALUE = _require.SCRIPT_VALUE_ALIASES_TO_VALUE;

describe('parser-unicode-properties', function () {
  it('general-category', function () {
    expect(parser.parse('/\\p{General_Category=Letter}/u').body).toEqual({
      type: 'UnicodeProperty',
      name: 'General_Category',
      value: 'Letter',
      negative: false,
      shorthand: false,
      binary: false,
      canonicalName: 'General_Category',
      canonicalValue: 'Letter'
    });
  });
  it('shorthand', function () {
    expect(parser.parse('/\\p{Letter}/u').body).toEqual({
      type: 'UnicodeProperty',
      name: 'General_Category',
      value: 'Letter',
      negative: false,
      shorthand: true,
      binary: false,
      canonicalName: 'General_Category',
      canonicalValue: 'Letter'
    });
  });
  it('negative', function () {
    expect(parser.parse('/\\P{Letter}/u').body).toEqual({
      type: 'UnicodeProperty',
      name: 'General_Category',
      value: 'Letter',
      negative: true,
      shorthand: true,
      binary: false,
      canonicalName: 'General_Category',
      canonicalValue: 'Letter'
    });
  });
  it('binary', function () {
    expect(parser.parse('/\\p{Hex_Digit}/u').body).toEqual({
      type: 'UnicodeProperty',
      name: 'Hex_Digit',
      value: 'Hex_Digit',
      negative: false,
      shorthand: false,
      binary: true,
      canonicalName: 'Hex_Digit',
      canonicalValue: 'Hex_Digit'
    });
  });
  it('script', function () {
    expect(parser.parse('/\\p{Script=Cyrillic}/u').body).toEqual({
      type: 'UnicodeProperty',
      name: 'Script',
      value: 'Cyrillic',
      negative: false,
      shorthand: false,
      binary: false,
      canonicalName: 'Script',
      canonicalValue: 'Cyrillic'
    });
  });
  it('script-extensions', function () {
    expect(parser.parse('/\\p{Script_Extensions=Cyrillic}/u').body).toEqual({
      type: 'UnicodeProperty',
      name: 'Script_Extensions',
      value: 'Cyrillic',
      negative: false,
      shorthand: false,
      binary: false,
      canonicalName: 'Script_Extensions',
      canonicalValue: 'Cyrillic'
    });
  });
  it('auto-general-category', function () {
    for (var value in GENERAL_CATEGORY_VALUE_TO_ALIASES) {
      expect(parser.parse("/\\p{General_Category=".concat(value, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'General_Category',
        value: value,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'General_Category',
        canonicalValue: value
      });
      expect(parser.parse("/\\p{gc=".concat(value, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'gc',
        value: value,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'General_Category',
        canonicalValue: value
      });
    }

    for (var _value in GENERAL_CATEGORY_VALUE_ALIASES_TO_VALUES) {
      expect(parser.parse("/\\p{General_Category=".concat(_value, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'General_Category',
        value: _value,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'General_Category',
        canonicalValue: getCanonicalValue(_value)
      });
      expect(parser.parse("/\\p{gc=".concat(_value, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'gc',
        value: _value,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'General_Category',
        canonicalValue: getCanonicalValue(_value)
      });
    }
  });
  it('auto-script', function () {
    for (var value in SCRIPT_VALUE_TO_ALIASES) {
      expect(parser.parse("/\\p{Script=".concat(value, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'Script',
        value: value,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'Script',
        canonicalValue: value
      });
      expect(parser.parse("/\\p{sc=".concat(value, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'sc',
        value: value,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'Script',
        canonicalValue: value
      });
    }

    for (var _value2 in SCRIPT_VALUE_ALIASES_TO_VALUE) {
      expect(parser.parse("/\\p{Script=".concat(_value2, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'Script',
        value: _value2,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'Script',
        canonicalValue: getCanonicalValue(_value2)
      });
      expect(parser.parse("/\\p{sc=".concat(_value2, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'sc',
        value: _value2,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'Script',
        canonicalValue: getCanonicalValue(_value2)
      });
    }
  });
  it('auto-script-extensions', function () {
    for (var value in SCRIPT_VALUE_TO_ALIASES) {
      expect(parser.parse("/\\p{Script_Extensions=".concat(value, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'Script_Extensions',
        value: value,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'Script_Extensions',
        canonicalValue: value
      });
      expect(parser.parse("/\\p{scx=".concat(value, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'scx',
        value: value,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'Script_Extensions',
        canonicalValue: value
      });
    }

    for (var _value3 in SCRIPT_VALUE_ALIASES_TO_VALUE) {
      expect(parser.parse("/\\p{Script_Extensions=".concat(_value3, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'Script_Extensions',
        value: _value3,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'Script_Extensions',
        canonicalValue: getCanonicalValue(_value3)
      });
      expect(parser.parse("/\\p{scx=".concat(_value3, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'scx',
        value: _value3,
        negative: false,
        shorthand: false,
        binary: false,
        canonicalName: 'Script_Extensions',
        canonicalValue: getCanonicalValue(_value3)
      });
    }
  });
  it('auto-shorthand', function () {
    for (var value in GENERAL_CATEGORY_VALUE_TO_ALIASES) {
      expect(parser.parse("/\\p{".concat(value, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'General_Category',
        value: value,
        negative: false,
        shorthand: true,
        binary: false,
        canonicalName: 'General_Category',
        canonicalValue: value
      });
    }

    for (var _value4 in GENERAL_CATEGORY_VALUE_ALIASES_TO_VALUES) {
      expect(parser.parse("/\\p{".concat(_value4, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: 'General_Category',
        value: _value4,
        negative: false,
        shorthand: true,
        binary: false,
        canonicalName: 'General_Category',
        canonicalValue: getCanonicalValue(_value4)
      });
    }
  });
  it('auto-binary', function () {
    for (var value in BINARY_PROP_NAMES_TO_ALIASES) {
      expect(parser.parse("/\\p{".concat(value, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: value,
        value: value,
        negative: false,
        shorthand: false,
        binary: true,
        canonicalName: value,
        canonicalValue: value
      });
    }

    for (var _value5 in BINARY_ALIASES_TO_PROP_NAMES) {
      expect(parser.parse("/\\p{".concat(_value5, "}/u")).body).toEqual({
        type: 'UnicodeProperty',
        name: _value5,
        value: _value5,
        negative: false,
        shorthand: false,
        binary: true,
        canonicalName: getCanonicalValue(_value5),
        canonicalValue: getCanonicalValue(_value5)
      });
    }
  });
});