"use strict";

/**
 * The MIT License (MIT)
 * Copyright (c) 2017-present Dmitry Soshnikov <dmitry.soshnikov@gmail.com>
 */
var traverse = require('..');

var parser = require('../../parser');

var defaultAst = parser.parse('/[a-z]+a/i');
describe('traverse-basic', function () {
  it('visits each node as a NodePath', function () {
    var visited = [];
    traverse.traverse(defaultAst, {
      RegExp: function RegExp(_ref) {
        var node = _ref.node,
            parent = _ref.parent;
        visited.push(node.type);
        expect(node.type).toBe('RegExp');
        expect(parent).toBe(null);
      },
      Alternative: function Alternative(_ref2) {
        var node = _ref2.node,
            parent = _ref2.parent;
        visited.push(node.type);
        expect(node.type).toBe('Alternative');
        expect(parent.type).toBe('RegExp');
        expect(node.expressions.length).toBe(2);
        expect(node.expressions[0].type).toBe('Repetition');
        expect(node.expressions[1].type).toBe('Char');
      },
      Repetition: function Repetition(_ref3) {
        var node = _ref3.node,
            parent = _ref3.parent;
        visited.push(node.type);
        expect(node.type).toBe('Repetition');
        expect(parent.type).toBe('Alternative');
        expect(node.expression.type).toBe('CharacterClass');
        expect(node.quantifier.type).toBe('Quantifier');
        expect(node.quantifier.kind).toBe('+');
      },
      CharacterClass: function CharacterClass(_ref4) {
        var node = _ref4.node,
            parent = _ref4.parent;
        visited.push(node.type);
        expect(node.type).toBe('CharacterClass');
        expect(parent.type).toBe('Repetition');
        expect(node.expressions.length).toBe(1);
        expect(node.expressions[0].type).toBe('ClassRange');
      },
      ClassRange: function ClassRange(_ref5) {
        var node = _ref5.node,
            parent = _ref5.parent,
            property = _ref5.property,
            index = _ref5.index;
        visited.push(node.type);
        expect(node.type).toBe('ClassRange');
        expect(parent.type).toBe('CharacterClass');
        expect(node.from.type).toBe('Char');
        expect(node.to.type).toBe('Char');
        expect(property).toBe('expressions');
        expect(index).toBe(0);
      },
      Quantifier: function Quantifier(_ref6) {
        var node = _ref6.node;
        visited.push(node.type);
        expect(node.type).toBe('Quantifier');
        expect(node.kind).toBe('+');
      },
      Char: function Char(_ref7) {
        var node = _ref7.node,
            parent = _ref7.parent;
        visited.push(node.type);
        expect(node.type).toBe('Char'); // Char appears only in these two parent nodes:

        expect(parent.type === 'ClassRange' || parent.type === 'Alternative').toBe(true);
      }
    }); // Visit order.

    expect(visited).toEqual(['RegExp', 'Alternative', 'Repetition', 'CharacterClass', 'ClassRange', 'Char', 'Char', 'Quantifier', 'Char']);
  });
  it('visits each node as raw', function () {
    var visited = []; // This traversal receives raw nodes in handlers.

    var options = {
      asNodes: true
    };
    traverse.traverse(defaultAst, {
      RegExp: function RegExp(node, parent) {
        visited.push(node.type);
        expect(node.type).toBe('RegExp');
        expect(parent).toBe(null);
      },
      Alternative: function Alternative(node, parent) {
        visited.push(node.type);
        expect(node.type).toBe('Alternative');
        expect(parent.type).toBe('RegExp');
        expect(node.expressions.length).toBe(2);
        expect(node.expressions[0].type).toBe('Repetition');
        expect(node.expressions[1].type).toBe('Char');
      },
      Repetition: function Repetition(node, parent) {
        visited.push(node.type);
        expect(node.type).toBe('Repetition');
        expect(parent.type).toBe('Alternative');
        expect(node.expression.type).toBe('CharacterClass');
        expect(node.quantifier.type).toBe('Quantifier');
        expect(node.quantifier.kind).toBe('+');
      },
      CharacterClass: function CharacterClass(node, parent) {
        visited.push(node.type);
        expect(node.type).toBe('CharacterClass');
        expect(parent.type).toBe('Repetition');
        expect(node.expressions.length).toBe(1);
        expect(node.expressions[0].type).toBe('ClassRange');
      },
      ClassRange: function ClassRange(node, parent, property, index) {
        visited.push(node.type);
        expect(node.type).toBe('ClassRange');
        expect(parent.type).toBe('CharacterClass');
        expect(node.from.type).toBe('Char');
        expect(node.to.type).toBe('Char');
        expect(property).toBe('expressions');
        expect(index).toBe(0);
      },
      Quantifier: function Quantifier(node) {
        visited.push(node.type);
        expect(node.type).toBe('Quantifier');
        expect(node.kind).toBe('+');
      },
      Char: function Char(node, parent) {
        visited.push(node.type);
        expect(node.type).toBe('Char'); // Char appears only in these two parent nodes:

        expect(parent.type === 'ClassRange' || parent.type === 'Alternative').toBe(true);
      }
    }, options); // Visit order.

    expect(visited).toEqual(['RegExp', 'Alternative', 'Repetition', 'CharacterClass', 'ClassRange', 'Char', 'Char', 'Quantifier', 'Char']);
  });
  it('modifies a direct node', function () {
    var ast = parser.parse('/a{1,}/');
    traverse.traverse(ast, {
      Quantifier: function Quantifier(_ref8) {
        var node = _ref8.node;

        if (node.kind === 'Range' && node.from == 1 && !node.to) {
          node.kind = '+';
          delete node.from;
        }
      }
    });
    expect(ast.body.quantifier).toEqual({
      type: 'Quantifier',
      kind: '+',
      greedy: true
    });
  });
  it('replaces a node using parent', function () {
    var ast = parser.parse('/a{1,}?/');
    traverse.traverse(ast, {
      Quantifier: function Quantifier(_ref9) {
        var node = _ref9.node,
            parent = _ref9.parent,
            property = _ref9.property;

        if (node.kind === 'Range' && node.from == 1 && !node.to) {
          parent[property] = {
            type: 'Quantifier',
            kind: '+',
            greedy: node.greedy
          };
        }
      }
    });
    expect(ast.body.quantifier).toEqual({
      type: 'Quantifier',
      kind: '+',
      greedy: false
    });
  });
  it('multiple handlers', function () {
    var ast = parser.parse('/a/');
    expect(ast.body.value).toBe('a'); // Two handlers.

    var handlers = [{
      Char: function Char(_ref10) {
        var node = _ref10.node;
        node.value = 'b';
      }
    }, {
      Char: function Char(_ref11) {
        var node = _ref11.node;
        node.value += 'c';
      }
    }];
    traverse.traverse(ast, handlers);
    expect(ast.body.value).toBe('bc');
  });
  it('shouldRun hook', function () {
    var handler = {
      shouldRun: function shouldRun(ast) {
        return ast.flags.includes('s');
      },
      RegExp: function RegExp(_ref12) {
        var node = _ref12.node;
        node.flags = node.flags.replace('s', '');
      },
      Char: function Char(path) {
        var node = path.node;

        if (node.kind === 'meta' && node.value === '.') {
          path.replace({
            type: 'CharacterClass',
            negative: true
          });
        }
      }
    }; // Should run.

    var ast = parser.parse('/./s');
    traverse.traverse(ast, handler);
    expect(ast).toEqual({
      type: 'RegExp',
      body: {
        type: 'CharacterClass',
        negative: true
      },
      flags: ''
    }); // Should not run.

    ast = parser.parse('/./');
    traverse.traverse(ast, handler);
    expect(ast).toEqual({
      type: 'RegExp',
      body: {
        type: 'Char',
        value: '.',
        symbol: '.',
        kind: 'meta',
        codePoint: NaN
      },
      flags: ''
    }); // Should run (not `shouldRun` hook present).

    ast = parser.parse('/./');
    traverse.traverse(ast, {
      Char: function Char(_ref13) {
        var node = _ref13.node,
            parent = _ref13.parent,
            property = _ref13.property;

        if (node.kind === 'meta' && node.value === '.') {
          parent[property] = {
            type: 'CharacterClass',
            negative: true
          };
        }
      }
    });
    expect(ast).toEqual({
      type: 'RegExp',
      body: {
        type: 'CharacterClass',
        negative: true
      },
      flags: ''
    });
  });
  it('catch-all * handler', function () {
    var ast = parser.parse('/a+/');
    var visited = [];
    traverse.traverse(ast, {
      '*': function _(_ref14) {
        var node = _ref14.node;
        visited.push(node.type);
      }
    });
    expect(visited).toEqual(['RegExp', 'Repetition', 'Char', 'Quantifier']);
  });
});